@@ -1,14 +1,27 @@
 /*
- * Copyright (c) 1997,2007 Andrew G Morgan <morgan@kernel.org>
+ * Copyright (c) 1997,2007,2016 Andrew G Morgan <morgan@kernel.org>
  *
  * This file deals with setting capabilities on files.
  */
 
 #include <sys/types.h>
-#include <attr/xattr.h>
 #include <byteswap.h>
+#include <sys/stat.h>
+#include <unistd.h>
+#include <linux/xattr.h>
 
-#define XATTR_SECURITY_PREFIX "security."
+/*
+ * We hardcode the prototypes for the Linux system calls here since
+ * there are no libcap library APIs that expose the user to these
+ * details, and that way we don't need to foce clients to link any
+ * other libraries to access them.
+ */
+extern ssize_t getxattr(const char *, const char *, void *, size_t);
+extern ssize_t fgetxattr(int, const char *, void *, size_t);
+extern int setxattr(const char *, const char *, const void *, size_t, int);
+extern int fsetxattr(int, const char *, const void *, size_t, int);
+extern int removexattr(const char *, const char *);
+extern int fremovexattr(int, const char *);
 
 #include "libcap.h"
 
@@ -82,7 +95,8 @@ static cap_t _fcaps_load(struct vfs_cap_data *rawvfscap, cap_t result,
     return result;
 }
 
-static int _fcaps_save(struct vfs_cap_data *rawvfscap, cap_t cap_d)
+static int _fcaps_save(struct vfs_cap_data *rawvfscap, cap_t cap_d,
+		       int *bytes_p)
 {
     __u32 eff_not_zero, magic;
     unsigned tocopy, i;
@@ -97,6 +111,7 @@ static int _fcaps_save(struct vfs_cap_data *rawvfscap, cap_t cap_d)
     case _LINUX_CAPABILITY_VERSION_1:
 	magic = VFS_CAP_REVISION_1;
 	tocopy = VFS_CAP_U32_1;
+	*bytes_p = XATTR_CAPS_SZ_1;
 	break;
 #endif
 
@@ -104,6 +119,15 @@ static int _fcaps_save(struct vfs_cap_data *rawvfscap, cap_t cap_d)
     case _LINUX_CAPABILITY_VERSION_2:
 	magic = VFS_CAP_REVISION_2;
 	tocopy = VFS_CAP_U32_2;
+	*bytes_p = XATTR_CAPS_SZ_2;
+	break;
+#endif
+
+#ifdef _LINUX_CAPABILITY_VERSION_3
+    case _LINUX_CAPABILITY_VERSION_3:
+	magic = VFS_CAP_REVISION_2;
+	tocopy = VFS_CAP_U32_2;
+	*bytes_p = XATTR_CAPS_SZ_2;
 	break;
 #endif
 
@@ -174,7 +198,7 @@ cap_t cap_get_fd(int fildes)
 	/* fill the capability sets via a system call */
 	sizeofcaps = fgetxattr(fildes, XATTR_NAME_CAPS,
 			       &rawvfscap, sizeof(rawvfscap));
-	if (sizeofcaps < sizeof(rawvfscap.magic_etc)) {
+	if (sizeofcaps < ssizeof(rawvfscap.magic_etc)) {
 	    cap_free(result);
 	    result = NULL;
 	} else {
@@ -186,7 +210,7 @@ cap_t cap_get_fd(int fildes)
 }
 
 /*
- * Set the capabilities on a named file.
+ * Get the capabilities from a named file.
  */
 
 cap_t cap_get_file(const char *filename)
@@ -204,7 +228,7 @@ cap_t cap_get_file(const char *filename)
 	/* fill the capability sets via a system call */
 	sizeofcaps = getxattr(filename, XATTR_NAME_CAPS,
 			      &rawvfscap, sizeof(rawvfscap));
-	if (sizeofcaps < sizeof(rawvfscap.magic_etc)) {
+	if (sizeofcaps < ssizeof(rawvfscap.magic_etc)) {
 	    cap_free(result);
 	    result = NULL;
 	} else {
@@ -223,18 +247,29 @@ cap_t cap_get_file(const char *filename)
 int cap_set_fd(int fildes, cap_t cap_d)
 {
     struct vfs_cap_data rawvfscap;
+    int sizeofcaps;
+    struct stat buf;
+
+    if (fstat(fildes, &buf) != 0) {
+	_cap_debug("unable to stat file descriptor %d", fildes);
+	return -1;
+    }
+    if (S_ISLNK(buf.st_mode) || !S_ISREG(buf.st_mode)) {
+	_cap_debug("file descriptor %d for non-regular file", fildes);
+	errno = EINVAL;
+	return -1;
+    }
 
     if (cap_d == NULL) {
 	_cap_debug("deleting fildes capabilities");
 	return fremovexattr(fildes, XATTR_NAME_CAPS);
-    } else if (_fcaps_save(&rawvfscap, cap_d) != 0) {
+    } else if (_fcaps_save(&rawvfscap, cap_d, &sizeofcaps) != 0) {
 	return -1;
     }
 
     _cap_debug("setting fildes capabilities");
 
-    return fsetxattr(fildes, XATTR_NAME_CAPS, &rawvfscap,
-		     sizeof(rawvfscap), 0);
+    return fsetxattr(fildes, XATTR_NAME_CAPS, &rawvfscap, sizeofcaps, 0);
 }
 
 /*
@@ -244,17 +279,28 @@ int cap_set_fd(int fildes, cap_t cap_d)
 int cap_set_file(const char *filename, cap_t cap_d)
 {
     struct vfs_cap_data rawvfscap;
+    int sizeofcaps;
+    struct stat buf;
+
+    if (lstat(filename, &buf) != 0) {
+	_cap_debug("unable to stat file [%s]", filename);
+	return -1;
+    }
+    if (S_ISLNK(buf.st_mode) || !S_ISREG(buf.st_mode)) {
+	_cap_debug("file [%s] is not a regular file", filename);
+	errno = EINVAL;
+	return -1;
+    }
 
     if (cap_d == NULL) {
 	_cap_debug("removing filename capabilities");
 	return removexattr(filename, XATTR_NAME_CAPS);
-    } else if (_fcaps_save(&rawvfscap, cap_d) != 0) {
+    } else if (_fcaps_save(&rawvfscap, cap_d, &sizeofcaps) != 0) {
 	return -1;
     }
 
     _cap_debug("setting filename capabilities");
-    return setxattr(filename, XATTR_NAME_CAPS, &rawvfscap,
-		    sizeof(rawvfscap), 0);
+    return setxattr(filename, XATTR_NAME_CAPS, &rawvfscap, sizeofcaps, 0);
 }
 
 #else /* ie. ndef VFS_CAP_U32 */
